@@ -4,19 +4,9 @@
  *
  * Copyright (C) 2007  Kristian Hoegsberg <krh@bitplanet.net>
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software Foundation,
- * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * This library is licensed under the GNU Lesser General Public License (LGPL),
+ * version 2.1 or later. See the file COPYING.LIB in the distribution for
+ * details.
  */
 
 #include <stdlib.h>
@@ -125,17 +115,19 @@ scan_devices(fw_handle_t handle)
 	char filename[32];
 	struct fw_cdev_get_info get_info;
 	struct fw_cdev_event_bus_reset reset;
-	int fd, err, i;
+	int fd, err, i, j, fname_str_sz;
 	struct port *ports;
 
 	ports = handle->ports;
 	memset(ports, 0, sizeof handle->ports);
+	for (i = 0; i < MAX_PORTS; i++)
+		ports[i].card = -1;
+
 	dir = opendir(FW_DEVICE_DIR);
 	if (dir == NULL)
 		return -1;
 
-	i = 0;
-	while (1) {
+	for (i = 0; i < MAX_PORTS; ) {
 		de = readdir(dir);
 		if (de == NULL)
 			break;
@@ -161,9 +153,13 @@ scan_devices(fw_handle_t handle)
 		if (err < 0)
 			continue;
 
-		if (i < MAX_PORTS && reset.node_id == reset.local_node_id) {
-			strncpy(ports[i].device_file, filename,
-				sizeof ports[i].device_file);
+		for (j = 0; j < i; j++)
+			if (ports[j].card == get_info.card)
+				break;
+		if (j == i) {
+			fname_str_sz = sizeof(ports[i].device_file) - 1;
+			strncpy(ports[i].device_file, filename, fname_str_sz);
+			ports[i].device_file[fname_str_sz] = '\0';
 			ports[i].node_count = (reset.root_node_id & 0x3f) + 1;
 			ports[i].card = get_info.card;
 			i++;
@@ -315,7 +311,7 @@ handle_inotify(raw1394handle_t handle, struct epoll_closure *ec,
 	struct fw_cdev_get_info info;
 	struct fw_cdev_event_bus_reset reset;
 	struct epoll_event ep;
-	int i, len, fd, phy_id;
+	int i, len, fd, phy_id, fname_str_sz;
 
 	event = (struct inotify_event *) fwhandle->buffer;
 	len = read(fwhandle->inotify_fd, event, BUFFER_SIZE);
@@ -365,8 +361,9 @@ handle_inotify(raw1394handle_t handle, struct epoll_closure *ec,
 	fwhandle->devices[i].node_id = reset.node_id;
 	fwhandle->devices[i].generation = reset.generation;
 	fwhandle->devices[i].fd = fd;
-	strncpy(fwhandle->devices[i].filename, filename,
-		sizeof fwhandle->devices[i].filename);
+	fname_str_sz = sizeof(fwhandle->devices[i].filename) - 1;
+	strncpy(fwhandle->devices[i].filename, filename, fname_str_sz);
+	fwhandle->devices[i].filename[fname_str_sz] = '\0';
 	fwhandle->devices[i].closure.func = handle_device_event;
 	ep.events = EPOLLIN;
 	ep.data.ptr = &fwhandle->devices[i].closure;
@@ -501,8 +498,10 @@ fw_handle_t fw_new_handle_on_port(int port)
 	if (handle == NULL)
 		return NULL;
 
-	if (fw_set_port(handle, port) < 0)
+	if (fw_set_port(handle, port) < 0) {
+		fw_destroy_handle(handle);
 		return NULL;
+	}
 
 	return handle;
 }
@@ -538,15 +537,17 @@ int fw_get_port_info(fw_handle_t handle,
 			  struct raw1394_portinfo *pinf,
 			  int maxports)
 {
-	int i;
+	int i, port_name_sz;
 
 	if (maxports >= handle->port_count)
 		maxports = handle->port_count;
 
 	for (i = 0; i < maxports; i++) {
 		pinf[i].nodes = handle->ports[i].node_count;
+		port_name_sz = sizeof(pinf[i].name) - 1;
 		strncpy(pinf[i].name, handle->ports[i].device_file,
-			sizeof pinf[i].name);
+			port_name_sz);
+		pinf[i].name[port_name_sz] = '\0';
 	}
 
 	return handle->port_count;
@@ -560,7 +561,7 @@ int fw_set_port(fw_handle_t handle, int port)
 	struct dirent *de;
 	char filename[32];
 	DIR *dir;
-	int i, fd, phy_id;
+	int i, fd, phy_id, fname_str_sz;
 
 	if (port >= handle->port_count) {
 		errno = EINVAL;
@@ -606,8 +607,9 @@ int fw_set_port(fw_handle_t handle, int port)
 		handle->devices[i].node_id = reset.node_id;
 		handle->devices[i].generation = reset.generation;
 		handle->devices[i].fd = fd;
-		strncpy(handle->devices[i].filename, filename,
-			sizeof handle->devices[i].filename);
+		fname_str_sz = sizeof(handle->devices[i].filename) -1;
+		strncpy(handle->devices[i].filename, filename, fname_str_sz);
+		handle->devices[i].filename[fname_str_sz] = '\0';
 
 		handle->devices[i].closure.func = handle_device_event;
 		memset(&ep, 0, sizeof(ep));
@@ -619,19 +621,27 @@ int fw_set_port(fw_handle_t handle, int port)
 			return -1;
 		}
 
-		handle->generation = reset.generation;
-		if (reset.node_id == reset.local_node_id) {
+		if (handle->iso.filename == NULL) {
 			memcpy(&handle->reset, &reset, sizeof handle->reset);
-			handle->local_fd = fd;
-			strncpy(handle->local_filename, filename,
-				sizeof handle->local_filename);
+			handle->iso.filename = handle->devices[i].filename;
+			handle->ioctl_fd = fd;
 		}
 
+		if (reset.node_id == reset.local_node_id)
+			handle->local_device = &handle->devices[i];
+
+		handle->generation = reset.generation;
+
 		i++;
 	}
 
 	closedir(dir);
 
+	if (i == 0) {
+		errno = ENODEV;
+		return -1;
+	}
+
 	return 0;
 }
 
@@ -648,7 +658,7 @@ int fw_reset_bus_new(fw_handle_t handle, int type)
 		break;
 	}
 
-	return ioctl(handle->local_fd,
+	return ioctl(handle->ioctl_fd,
 		     FW_CDEV_IOC_INITIATE_BUS_RESET, &initiate);
 }
 
@@ -765,10 +775,12 @@ handle_arm_request(raw1394handle_t handle, struct address_closure *ac,
 	}
 	rrb->request.generation = fwhandle->reset.generation;
 	rrb->request.buffer_length = in_length;
+	rrb->request.buffer = rrb->data;
 	memcpy(rrb->request.buffer, request->data, in_length);
 
 	rrb->response.response_code = response.rcode;
 	rrb->response.buffer_length = response.length;
+	rrb->response.buffer = rrb->data + in_length;
 	memcpy(rrb->response.buffer,
 	       allocation->data + offset, response.length);
 
@@ -807,7 +819,7 @@ fw_arm_register(fw_handle_t handle, nodeaddr_t start,
 	request.length = length;
 	request.closure = ptr_to_u64(&allocation->closure);
 
-	retval = ioctl(handle->local_fd, FW_CDEV_IOC_ALLOCATE, &request);
+	retval = ioctl(handle->ioctl_fd, FW_CDEV_IOC_ALLOCATE, &request);
 	if (retval < 0) {
 		free(allocation);
 		return -1;
@@ -853,7 +865,7 @@ fw_arm_unregister(fw_handle_t handle, nodeaddr_t start)
 	request.handle = allocation->handle;
 	free(allocation);
 
-	return ioctl(handle->local_fd, FW_CDEV_IOC_DEALLOCATE, &request);
+	return ioctl(handle->ioctl_fd, FW_CDEV_IOC_DEALLOCATE, &request);
 }
 
 int
@@ -918,7 +930,8 @@ fw_start_phy_packet_write(fw_handle_t handle,
 static int
 send_request(fw_handle_t handle, int tcode,
 	     nodeid_t node, nodeaddr_t addr,
-	     size_t length, void *in, void *out, unsigned long tag)
+	     size_t in_length, void *in, size_t out_length, void *out,
+	     unsigned long tag)
 {
 	struct fw_cdev_send_request *request;
 	struct request_closure *closure;
@@ -951,14 +964,14 @@ send_request(fw_handle_t handle, int tcode,
 	}
 
 	closure->data = out;
-	closure->length = length;
+	closure->length = out_length;
 	closure->tag = tag;
 
 	request = (struct fw_cdev_send_request *) handle->buffer;
 	request->tcode = tcode;
 	request->generation = handle->generation;
 	request->offset = addr;
-	request->length = length;
+	request->length = in_length > out_length ? in_length : out_length;
 	request->closure = ptr_to_u64(closure);
 	request->data = ptr_to_u64(in);
 
@@ -977,7 +990,7 @@ fw_start_read(fw_handle_t handle, nodeid_t node, nodeaddr_t addr,
 		tcode = TCODE_READ_BLOCK_REQUEST;
 
 	return send_request(handle, tcode,
-			    node, addr, length, NULL, buffer, tag);
+			    node, addr, 0, NULL, length, buffer, tag);
 }
 
 int
@@ -992,7 +1005,7 @@ fw_start_write(fw_handle_t handle, nodeid_t node, nodeaddr_t addr,
 		tcode = TCODE_WRITE_BLOCK_REQUEST;
 
 	return send_request(handle, tcode,
-			    node, addr, length, data, NULL, tag);
+			    node, addr, length, data, 0, NULL, tag);
 }
 
 static int
@@ -1054,7 +1067,8 @@ fw_start_lock(fw_handle_t handle, nodeid_t node, nodeaddr_t addr,
 		return length;
 
 	return send_request(handle, 16 + extcode,
-			    node, addr, length, buffer, result, tag);
+			    node, addr, length, buffer,
+			    sizeof *result, result, tag);
 }
 
 int
@@ -1070,7 +1084,8 @@ fw_start_lock64(fw_handle_t handle, nodeid_t node, nodeaddr_t addr,
 		return length;
 
 	return send_request(handle, 16 + extcode,
-			    node, addr, length, buffer, result, tag);
+			    node, addr, length, buffer,
+			    sizeof *result, result, tag);
 }
 
 int
@@ -1080,6 +1095,7 @@ fw_start_async_stream(fw_handle_t handle, unsigned int channel,
 			   unsigned long rawtag)
 {
 	/* FIXME: implement this? */
+	errno = ENOSYS;
 	return -1;
 }
 
@@ -1091,6 +1107,7 @@ fw_start_async_send(fw_handle_t handle,
 			 quadlet_t *data, unsigned long rawtag)
 {
 	/* FIXME: implement this? */
+	errno = ENOSYS;
 	return -1;
 }
 
@@ -1113,7 +1130,7 @@ sync_callback(raw1394handle_t handle, void *data, raw1394_errcode_t err)
 static int
 send_request_sync(raw1394handle_t handle, int tcode,
 		  nodeid_t node, nodeaddr_t addr,
-		  size_t length, void *in, void *out)
+		  size_t in_length, void *in, size_t out_length, void *out)
 {
 	fw_handle_t fwhandle = handle->mode.fw;
 	struct raw1394_reqhandle reqhandle;
@@ -1124,7 +1141,8 @@ send_request_sync(raw1394handle_t handle, int tcode,
 	reqhandle.data = &sd;
 
 	err = send_request(fwhandle, tcode, node, addr,
-			   length, in, out, (unsigned long) &reqhandle);
+			   in_length, in, out_length, out,
+			   (unsigned long) &reqhandle);
 
 	while (!sd.done) {
 		if (err < 0)
@@ -1150,7 +1168,7 @@ fw_read(raw1394handle_t handle, nodeid_t node, nodeaddr_t addr,
 		tcode = TCODE_READ_BLOCK_REQUEST;
 
 	return send_request_sync(handle, tcode,
-				 node, addr, length, NULL, buffer);
+				 node, addr, 0, NULL, length, buffer);
 }
 
 int
@@ -1165,7 +1183,7 @@ fw_write(raw1394handle_t handle, nodeid_t node, nodeaddr_t addr,
 		tcode = TCODE_WRITE_BLOCK_REQUEST;
 
 	return send_request_sync(handle, tcode,
-				 node, addr, length, data, NULL);
+				 node, addr, length, data, 0, NULL);
 }
 
 int
@@ -1174,14 +1192,15 @@ fw_lock(raw1394handle_t handle, nodeid_t node, nodeaddr_t addr,
 	     quadlet_t *result)
 {
 	quadlet_t buffer[2];
-	size_t length;
+	ssize_t length;
 
 	length = setup_lock(extcode, data, arg, buffer);
 	if (length < 0)
 		return length;
 
 	return send_request_sync(handle, 16 + extcode, node, addr,
-				 length, buffer, result);
+				 (size_t) length, buffer,
+				 sizeof *result, result);
 }
 
 int
@@ -1190,14 +1209,15 @@ fw_lock64(raw1394handle_t handle, nodeid_t node, nodeaddr_t addr,
 	       octlet_t *result)
 {
 	octlet_t buffer[2];
-	size_t length;
+	ssize_t length;
 
 	length = setup_lock64(extcode, data, arg, buffer);
 	if (length < 0)
 		return length;
 
 	return send_request_sync(handle, 16 + extcode, node, addr,
-				 length, buffer, result);
+				 (size_t) length, buffer,
+				 sizeof *result, result);
 }
 
 int
@@ -1206,6 +1226,7 @@ fw_async_stream(fw_handle_t handle, unsigned int channel,
 		     size_t length, quadlet_t *data)
 {
 	/* FIXME: implement this? */
+	errno = ENOSYS;
 	return -1;
 }
 
@@ -1216,6 +1237,7 @@ fw_async_send(fw_handle_t handle,
 		   quadlet_t *data)
 {
 	/* FIXME: implement this? */
+	errno = ENOSYS;
 	return -1;
 }
 
@@ -1235,7 +1257,7 @@ fw_start_fcp_listen(fw_handle_t handle)
 	request.offset = CSR_REGISTER_BASE + CSR_FCP_COMMAND;
 	request.length = CSR_FCP_END - CSR_FCP_COMMAND;
 	request.closure = ptr_to_u64(closure);
-	if (ioctl(handle->local_fd, FW_CDEV_IOC_ALLOCATE, &request) < 0)
+	if (ioctl(handle->ioctl_fd, FW_CDEV_IOC_ALLOCATE, &request) < 0)
 		return -1;
 
 	handle->fcp_allocation_handle = request.handle;
@@ -1250,13 +1272,14 @@ fw_stop_fcp_listen(fw_handle_t handle)
 
 	request.handle = handle->fcp_allocation_handle;
 
-	return ioctl(handle->local_fd, FW_CDEV_IOC_DEALLOCATE, &request);
+	return ioctl(handle->ioctl_fd, FW_CDEV_IOC_DEALLOCATE, &request);
 }
 
 int
 fw_update_config_rom(fw_handle_t handle, const quadlet_t *new_rom,
 			  size_t size, unsigned char rom_version)
 {
+	errno = ENOSYS;
 	return -1;
 }
 
@@ -1268,13 +1291,18 @@ fw_get_config_rom(fw_handle_t handle, quadlet_t *buffer,
 	struct fw_cdev_get_info get_info;
 	int err;
 
+	if (handle->local_device == NULL) {
+		errno = EPERM;
+		return -1;
+	}
+
 	memset(&get_info, 0, sizeof(get_info));
 	get_info.version = FW_CDEV_VERSION;
 	get_info.rom = ptr_to_u64(buffer);
 	get_info.rom_length = buffersize;
 	get_info.bus_reset = 0;
 
-	err = ioctl(handle->local_fd, FW_CDEV_IOC_GET_INFO, &get_info);
+	err = ioctl(handle->local_device->fd, FW_CDEV_IOC_GET_INFO, &get_info);
 	if (err)
 		return err;
 
@@ -1283,100 +1311,3 @@ fw_get_config_rom(fw_handle_t handle, quadlet_t *buffer,
 
 	return 0;
 }
-
-#define MAXIMUM_BANDWIDTH 4915
-
-int
-fw_bandwidth_modify (raw1394handle_t handle,
-			  unsigned int bandwidth,
-			  enum raw1394_modify_mode mode)
-{
-	quadlet_t buffer, compare, swap;
-	nodeaddr_t addr;
-	int result;
-
-        if (bandwidth == 0)
-                return 0;
-
-	addr = CSR_REGISTER_BASE + CSR_BANDWIDTH_AVAILABLE;
-        /* Read current bandwidth usage from IRM. */
-        result = raw1394_read (handle, raw1394_get_irm_id (handle), addr,
-			       sizeof buffer, &buffer);
-        if (result < 0)
-                return -1;
-
-        compare = ntohl (buffer);
-	switch (mode) {
-	case RAW1394_MODIFY_ALLOC:
-		swap = compare - bandwidth;
-		if (swap < 0)
-			return -1;
-		break;
-
-	case RAW1394_MODIFY_FREE:
-		swap = compare + bandwidth;
-		if (swap > MAXIMUM_BANDWIDTH)
-			swap = MAXIMUM_BANDWIDTH;
-		break;
-
-	default:
-		return -1;
-	}
-
-	result = raw1394_lock(handle, raw1394_get_irm_id (handle), addr,
-			      RAW1394_EXTCODE_COMPARE_SWAP,
-			      htonl(swap), htonl(compare), &buffer);
-	if (result < 0 || ntohl(buffer) != compare)
-		return -1;
-  
-        return 0;
-}
-
-int
-fw_channel_modify (raw1394handle_t handle,
-			unsigned int channel,
-			enum raw1394_modify_mode mode)
-{
-        quadlet_t buffer, compare, swap, bit;
-        nodeaddr_t addr;
-        int result;
-        
-	if (channel >= 64)
-		return -1;
-	addr = CSR_REGISTER_BASE +
-		CSR_CHANNELS_AVAILABLE_HI + 4 * (channel / 32);
-	/* Read currently available channels from IRM. */
-        result = raw1394_read(handle, raw1394_get_irm_id (handle), addr, 
-			      sizeof buffer, &buffer);
-        if (result < 0)
-                return -1;
-        
-	/* IEEE numbers bits from MSB (0) to LSB (31). */
-        bit = 1 << (31 - (channel & 31));
-        compare = ntohl(buffer);
-	switch (mode) {
-	case RAW1394_MODIFY_ALLOC:
-                if ((compare & bit) == 0)
-                        return -1;
-                swap = buffer & ~bit;
-		break;
-
-        case RAW1394_MODIFY_FREE:
-                if ((buffer & bit) != 0)
-                        return -1;
-                swap = buffer | bit;
-		break;
-
-	default:
-		return -1;
-        }
-  
-        result = raw1394_lock (handle, raw1394_get_irm_id (handle), addr,
-			       RAW1394_EXTCODE_COMPARE_SWAP,
-			       htonl(swap), htonl(compare), &buffer);
-
-        if (result < 0 || ntohl(buffer) != compare)
-                return -1;
-  
-        return 0;
-}
